
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "xpadtrans.h"

static _keydef_node first = {
    "*default*", 1,
    {
        XK_Up, XK_Down, XK_Left, XK_Right,
        XK_space, XK_Shift_L,
        0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
    },
    {
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
    },
    NULL,
};

_keydef_node *keydef_list_search(const char *name)
{
    _keydef_node *n = &first;
    if (name == NULL) return &first;
    while (n){
        if (!strcmp(n->name, name))
            return n;
        n = n->next;
    }
    return NULL;
}

int keydef_list_add(_keydef_node *keydef_node)
{
    _keydef_node *n = &first;
#ifdef DEBUG
    fprintf(stderr, "keydef_list_add %s\n", keydef_node->name);
#endif
    for (; n->next; n=n->next);
    keydef_node->next = NULL;
    n->next = keydef_node;
    return 0;
}

static int keydef_list_read_line(FILE *fp, char *buffer)
{
    char c;
    int i;
    i = 0;
    while (1){
        c = getc(fp);
        if (c == EOF) return 1;
        if (c == '\n'){
            if (i) break;
        }
        else
            buffer[i++] = c;
        if (i >= 1023) break;
    }
    buffer[i] = 0;
    return 0;
}

static int keydef_list_load_one(FILE *fp)
{
    _keydef_node *n;
    char buffer[1024];
    int i, new = 0;
    if (keydef_list_read_line(fp, buffer)) return 1;
    if (!buffer[0]) return -1;
    n = keydef_list_search(buffer);
    if (!n){
        n = malloc(sizeof(_keydef_node));
        new = 1;
        if (!n) return -1;
        strcpy(n->name, buffer);
    }
    if (keydef_list_read_line(fp, buffer)) return -1;
    if (!buffer[0]) return -1;
    if (!strcmp(buffer, "send_to_focus"))
        n->send_to_focus = 1;
    else
        n->send_to_focus = 0;
    for (i=0; i<16; i++){
        if(fscanf(fp, "%s", buffer) == EOF) return -1;
        n->key[i] = XStringToKeysym(buffer);
        if(fscanf(fp, "%s", buffer) == EOF) return -1;
        n->key_alt[i] = XStringToKeysym(buffer);
    }
    if (new){
        keydef_list_add(n);
    }
    return 0;
}

int keydef_list_load(const char *path)
{
    FILE *fp;
    char path_resume[1024];
    int r;
#ifdef DEBUG
    fprintf(stderr, "keydef_list_load %s\n", path);
#endif
    sprintf(path_resume, "%s.resume", path);
    fp = fopen(path_resume, "r");
    if (fp != NULL){
        fclose(fp);
        fprintf(stderr, "resume file (%s) exists.\n", path_resume);
        return -1;
    }
    fp = fopen(path, "r");
    if (fp == NULL) return 0;
    while (!(r = keydef_list_load_one(fp)));
    fclose(fp);
    return (r < 0) ? 1 : 0;
}

static int keydef_list_diff(_keydef_node *n)
{
    int i;
    if (first.send_to_focus != n->send_to_focus) return 1;
    for (i=0; i<16; i++){
        if (first.key[i] != n->key[i]) return 1;
        if (first.key_alt[i] != n->key_alt[i]) return 1;
    }
    return 0;
}

int keydef_list_save(const char *path)
{
    FILE *fp;
    char path_resume[1024];
    _keydef_node *n = &first;
    int i;
#ifdef DEBUG
    fprintf(stderr, "keydef_list_save %s\n", path);
#endif
    sprintf(path_resume, "%s.resume", path);
    if (link(path, path_resume) < 0){
        fprintf(stderr, "failed to link(%s, %s).\n", path, path_resume);
        return -1;
    }
    if (unlink(path)){
        fprintf(stderr, "failed to unlink(%s).\n", path);
    }
    fp = fopen(path, "w");
    if (fp == NULL) return 1;
    while (n){
        if ((n != &first) && !keydef_list_diff(n)){
            n = n->next;
            continue;
        }
        fprintf(fp, "\n%s\n", n->name);
        fprintf(fp, "%s\n",
                (n->send_to_focus) ? "send_to_focus" : "send_to_pointer");
        for (i=0; i<16; i++){
            fprintf(fp, "%s %s\n",
                    XKeysymToString(n->key[i]),
                    XKeysymToString(n->key_alt[i]));
        }
#ifdef DEBUG
        fprintf(stderr, "\n%s\n", n->name);
        fprintf(stderr, "%s\n",
                (n->send_to_focus) ? "send_to_focus" : "send_to_pointer");
        for (i=0; i<16; i++){
            fprintf(stderr, "%s %s\n",
                    XKeysymToString(n->key[i]),
                    XKeysymToString(n->key_alt[i]));
        }
#endif
        n = n->next;
    }
    fclose(fp);
    if (unlink(path_resume)){
        fprintf(stderr, "failed to unlink(%s).\n", path_resume);
    }
    return 0;
}
